import unittest
import flavio


class TestTauLGamma(unittest.TestCase):
    def test_taulgamma(self):
        # compare to the experimental values
        self.assertEqual(flavio.sm_prediction('BR(tau->mugamma)'), 0)
        self.assertEqual(flavio.sm_prediction('BR(tau->egamma)'), 0)



r"""Functions for the lepton flavour violating $\mu\to 3e$ decay."""

import flavio
from math import pi, sqrt

from flavio.physics.taudecays.tau3l import _BR_tau3mu


def BR_mu3e(wc_obj, par):
    r"""Branching ratio of $\mu^-\to e^-e^+e^-$."""
    scale = flavio.config['renormalization scale']['mudecays']
    wc = wc_obj.get_wc('mue', scale, par, nf_out=4)
    alpha = flavio.physics.running.running.get_alpha_e(par, scale, nf_out=3)
    e = sqrt(4 * pi * alpha)
    # cf. (22, 23) of hep-ph/0404211
    pre_br = 1 / (8 * par['GF']**2)
    mmu = par['m_mu']
    me = par['m_e']
    pre_wc_1 = 1 / e / mmu
    pre_wc_2 = 1
    DLg = pre_wc_1 * wc['Cgamma_mue']
    DRg = pre_wc_1 * wc['Cgamma_emu'].conjugate()
    FLL = pre_wc_2 * wc['CVLL_eemue']
    FLR = pre_wc_2 * wc['CVLR_eemue']
    FRL = pre_wc_2 * wc['CVLR_mueee']
    FRR = pre_wc_2 * wc['CVRR_eemue']
    SRR = pre_wc_2 * wc['CSRR_eemue']
    SLL = pre_wc_2 * wc['CSRR_eeemu'].conjugate()
    br_wc = _BR_tau3mu(mmu, me, FLL, FLR, FRL, FRR, DLg, DRg, SRR, SLL)
    return pre_br * br_wc


_process_tex = r"\mu^-\to e^-e^+e^-"
_process_taxonomy = r'Process :: muon decays :: LFV decays :: $' + _process_tex + r"$"

_obs_name = "BR(mu->eee)"
_obs = flavio.classes.Observable(_obs_name)
_obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
_obs.tex = r"$\text{BR}(" + _process_tex + r")$"
_obs.add_taxonomy(_process_taxonomy)
flavio.classes.Prediction(_obs_name, BR_mu3e)



r"""Functions for lepton flavour violating $\tau\to \ell_1\ell_2\ell_2$ decays."""

import flavio
from math import pi, log, sqrt


# names of LFV sectors in WCxf
wcxf_sector_names = {('tau', 'mu'): 'mutau',
                     ('tau', 'e'): 'taue',
                     ('mu', 'e'): 'mue', }


def _BR_taumuee(mtau, me, FLL, FLR, FRL, FRR, e2DLg, e2DRg, SRR, SLL):
    # (22) of hep-ph/0404211
    # FIXME scalar contributions missing!
    return (abs(FLL)**2 + abs(FLR)**2 + abs(FRL)**2 + abs(FRR)**2
            + 4 * (e2DLg * (FLL + FLR).conjugate()
                          + e2DRg * (FRL + FRR).conjugate()).real
            + 8 * (abs(e2DLg)**2 + abs(e2DRg)**2)
            * (log(mtau**2 / me**2) - 3))


def _BR_tau3mu(mtau, mmu, FLL, FLR, FRL, FRR, e2DLg, e2DRg, SRR, SLL):
    # (23) of hep-ph/0404211
    # (117) of hep-ph/9909265
    return (2 * abs(FLL)**2 + abs(FLR)**2 + abs(FRL)**2 + 2 * abs(FRR)**2
            + 1 / 8 * (abs(SLL)**2 + abs(SRR)**2)
            + 4 * (e2DLg * (2 * FLL + FLR).conjugate()
                          + e2DRg * (FRL + 2 * FRR).conjugate()).real
            + 8 * (abs(e2DLg)**2 + abs(e2DRg)**2)
            * (log(mtau**2 / mmu**2) - 11 / 4))


def BR_taumull(wc_obj, par, lep):
    r"""Branching ratio of $\tau^-\to\mu^-\ell^+\ell^-$."""
    scale = flavio.config['renormalization scale']['taudecays']
    sector = wcxf_sector_names['tau', 'mu']
    wc = wc_obj.get_wc(sector, scale, par, nf_out=4)
    alpha = flavio.physics.running.running.get_alpha_e(par, scale, nf_out=4)
    e = sqrt(4 * pi * alpha)
    # cf. (22, 23) of hep-ph/0404211
    pre_br = par['BR(tau->mununu)'] / (8 * par['GF']**2)
    mtau = par['m_tau']
    pre_wc_1 = 1 / e / mtau
    pre_wc_2 = 1
    e2DLg = e**2 * pre_wc_1 * wc['Cgamma_taumu']
    e2DRg = e**2 * pre_wc_1 * wc['Cgamma_mutau'].conjugate()
    FLL = pre_wc_2 * wc['CVLL_{}taumu'.format(2 * lep)]
    FLR = pre_wc_2 * wc['CVLR_{}taumu'.format(2 * lep)]
    FRL = pre_wc_2 * wc['CVLR_taumu{}'.format(2 * lep)]
    FRR = pre_wc_2 * wc['CVRR_{}taumu'.format(2 * lep)]
    SRR = pre_wc_2 * wc['CSRR_{}taumu'.format(2 * lep)]
    SLL = pre_wc_2 * wc['CSRR_{}mutau'.format(2 * lep)].conjugate()
    if lep == 'e':
        me = par['m_e']
        br_wc = _BR_taumuee(mtau, me, FLL, FLR, FRL, FRR, e2DLg, e2DRg, SRR, SLL)
    elif lep == 'mu':
        mmu = par['m_mu']
        br_wc = _BR_tau3mu(mtau, mmu, FLL, FLR, FRL, FRR, e2DLg, e2DRg, SRR, SLL)
    return pre_br * br_wc


# function returning function needed for prediction instance
def br_taumull_fct(lep):
    def f(wc_obj, par):
        return BR_taumull(wc_obj, par, lep)
    return f


# Observable and Prediction instances
_tex = {'e': 'e', 'mu': r'\mu'}

for lep in _tex:
    _process_tex = r"\tau^-\to \mu^-" + _tex[lep] + r"^+" + _tex[lep] + r"^-"
    _process_taxonomy = r'Process :: $\tau$ lepton decays :: LFV decays :: $\tau\to \ell^\prime\ell\ell$ :: $' + _process_tex + r"$"

    _obs_name = "BR(tau->mu" + 2 * lep + ")"
    _obs = flavio.classes.Observable(_obs_name)
    _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    flavio.classes.Prediction(_obs_name, br_taumull_fct(lep))



"""Generic formulae for fermion decay rates."""

from flavio.physics.common import lambda_K
from math import sqrt, pi

def GammaFvf(M, mv, mf, gL, gR):
    """Generic decay width of a fermion to a vector boson and a fermion."""
    amp2 = ((abs(gL)**2 + abs(gR)**2)
            * ((mf**2 - M**2)**2 /mv**2 + mf**2 + M**2 - 2. *mv**2)
            - 12 * mf * M * (gL.real * gR.real + gL.imag * gR.imag))/2.
    return Gamma(amp2, M, mf, mv)

def GammaFsf(M, ms, mf, gL, gR):
    """Generic decay width of a fermion to a scalar and a fermion."""
    amp2 = ((abs(gL)**2 + abs(gR)**2) * (M**2 + mf**2 - ms**2)
            + 4 * M * mf *(gL.real * gR.real + gL.imag * gR.imag))/2.
    return Gamma(amp2, M, mf, ms)

def Gamma(amp2, M, m1, m2):
    return sqrt(lambda_K(M**2,m1**2,m2**2)) /(16 * pi * M**3) * amp2



import unittest
import flavio


class TestTau3l(unittest.TestCase):
    def test_tau3l(self):
        # compare to the experimental values
        self.assertEqual(flavio.sm_prediction('BR(tau->muee)'), 0)
        self.assertEqual(flavio.sm_prediction('BR(tau->mumumu)'), 0)



r"""Electric dipole moment of the neutron."""

import flavio
from flavio.config import config
from math import pi, sqrt


def nedm_wceff(wc_obj, par):
    scale = config['renormalization scale']['nEDM']
    v2 = 1 / (sqrt(2) * par['GF'])
    md = flavio.physics.running.running.get_md(par, scale, nf_out=3)
    mu = flavio.physics.running.running.get_mu(par, scale, nf_out=3)
    alpha = flavio.physics.running.running.get_alpha(par, scale, nf_out=3)
    e = sqrt(4 * pi * alpha['alpha_e'])
    gs = sqrt(4 * pi * alpha['alpha_s'])
    wc = wc_obj.get_wc('dF=0', scale=2, par=par, eft='WET-3', basis='flavio')
    wceff = {}
    pre = 4 * par['GF'] / sqrt(2) / (8 * pi**2)
    wceff['edm_d'] = pre * e * md * wc['C7_dd'].imag
    wceff['edm_u'] = pre * e * mu * wc['C7_uu'].imag
    wceff['cedm_d'] = pre * gs * md * wc['C8_dd'].imag
    wceff['cedm_u'] = pre * gs * mu * wc['C8_uu'].imag
    wceff['Gtilde'] = wc['CGtilde']
    return wceff


def nedm(wc_obj, par):
    wceff = nedm_wceff(wc_obj, par)
    d = 0
    # quark EDM & CEDM contributions
    d += par['nEDM gT_d'] * wceff['edm_d']
    d += par['nEDM gT_u'] * wceff['edm_u']
    d += par['nEDM ~rho_d'] * wceff['cedm_d']
    d += par['nEDM ~rho_u'] * wceff['cedm_u']
    d += par['nEDM beta_G'] * wceff['Gtilde']
    return abs(d)


# Observable and Prediction instances

_obs_name = "d_n"
_obs = flavio.classes.Observable(name=_obs_name)
_obs.set_description(r"Electric dipole moment of the neutron")
_obs.tex = r"$d_n$"
_obs.add_taxonomy(r'Process :: Electric dipole moments :: Nucleons :: $d_n$')
flavio.classes.Prediction(_obs_name, nedm)



r"""Functions for $\tau\to V\ell$."""

import flavio
from math import sqrt, pi


# names of LFV sectors in WCxf
wcxf_sector_names = {('tau', 'mu'): 'mutau',
                     ('tau', 'e'): 'taue',
                     ('mu', 'e'): 'mue', }


def br_tauvl(wc_obj, par, V, lep):
    r"""Branching ratio of $\tau^+\to V^0\ell^+$."""
    scale = flavio.config['renormalization scale']['taudecays']
    sec = wcxf_sector_names['tau', lep]
    wc = wc_obj.get_wc(sec, scale, par, nf_out=4)
    alpha = flavio.physics.running.running.get_alpha_e(par, scale, nf_out=3)
    e = sqrt(4 * pi * alpha)
    mV = par['m_' + V]
    mtau = par['m_tau']
    x = mV**2 / mtau**2
    pre_wc_1 = 1 / e / mtau
    e2DLg = e**2 * pre_wc_1 * wc['Cgamma_tau{}'.format(lep)]
    e2DRg = e**2 * pre_wc_1 * wc['Cgamma_{}tau'.format(lep)].conjugate()
    F = {}
    for q in 'ud':
        F['LL' + q] = wc['CVLL_tau{}{}'.format(lep, 2 * q   )]
        F['RR' + q] = wc['CVRR_tau{}{}'.format(lep, 2 * q)]
        F['LR' + q] = wc['CVLR_tau{}{}'.format(lep, 2 * q)]
        F['RL' + q] = wc['CVLR_{}tau{}'.format(2 * q, lep)]
    if V == 'rho0':
        FL = (F['LLu'] - F['LLd']) / 2 + (F['LRu'] - F['LRd']) / 2
        FR = (F['RLu'] - F['RLd']) / 2 + (F['RRu'] - F['RRd']) / 2
        Vud = flavio.physics.ckm.get_ckm(par)[0, 0]
        norm = 1 / (4 * par['GF']**2 * abs(Vud)**2) * par['BR(tau->rhonu)']
    rWC = (abs(FL)**2 + abs(FR)**2
           - 6 / (1 + 2 * x) * (e2DLg * FL.conjugate() + e2DRg * FR.conjugate()).real
           + (2 + x) / (x * (1 + 2 * x)) * (abs(e2DLg)**2 + abs(e2DRg)**2))
    return norm * rWC


# function returning function needed for prediction instance
def br_tauvl_fct(V, lep):
    def f(wc_obj, par):
        return br_tauvl(wc_obj, par, V, lep)
    return f

# Observable and Prediction instances

_had = {'rho0': r'\rho^0',}
_shortname = {'rho0': 'rho',}
_lep = {'e': ' e', 'mu': r'\mu',}

for V in _had:
    for l in _lep:
        _obs_name = "BR(tau->" + _shortname[V] + l + r")"
        _obs = flavio.classes.Observable(_obs_name)
        _process_tex = r"\tau^+\to " + _had[V] + _lep[l] + r"^+"
        _process_taxonomy = r'Process :: $\tau$ lepton decays :: LFV decays :: $\tau\to V\ell$ :: $' + _process_tex + r"$"
        _obs.add_taxonomy(_process_taxonomy)
        _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
        flavio.classes.Prediction(_obs_name, br_tauvl_fct(V, l))
